from ..lookup_tools.jme_standard_function import jme_standard_function
import warnings
import re
import numpy as np
from copy import deepcopy
from awkward import JaggedArray

def _checkConsistency(against,tocheck):
    if against is None:
        against = tocheck
    else:
        if against != tocheck:
            raise Exception('Corrector for {} is mixed'/
                            'with correctors for {}!'.format(tocheck,against))
    return tocheck

_levelre = re.compile('Resolution')
def _getLevel(levelName):
    matches = _levelre.findall(levelName)
    if len(matches) > 1:
        raise Exception('Malformed JEC level name: {}'.format(levelName))
    return matches[0]

_level_order = ['Resolution']

class JetResolution(object):
    """
        This class is a columnar implementation of the JetResolution tool in
        CMSSW and FWLite. It calculates the jet energy resolution for a corrected jet
        in a given binning.
        You can use this class as follows:
        jr = JetResolution(name1=corrL1,...)
        jetRes = jr(JetParameter1=jet.parameter1,...)
    """
    def __init__(self,**kwargs):
        """
            You construct a JetResolution by passing in a dict of names and functions.
            Names must be formatted as '<campaign>_<dataera>_<datatype>_<level>_<jettype>'.
        """
        jettype = None
        levels = []
        funcs = []
        datatype = None
        campaign = None
        dataera = None
        for name,func in kwargs.items():
            if not isinstance(func,jme_standard_function):
                raise Exception('{} is a {} and not a jme_standard_function!'.format(name,
                                                                                     type(func)))
            info = name.split('_')
            if len(info) != 5:
                raise Exception('Corrector name is not properly formatted!')
            
            campaign = _checkConsistency(campaign,info[0])
            dataera  = _checkConsistency(dataera,info[1])
            datatype = _checkConsistency(datatype,info[2])
            levels.append(info[3])
            funcs.append(func)
            jettype  = _checkConsistency(jettype,info[4])
        
        if campaign is None:
            raise Exception('Unable to determine production campaign of JECs!')
        else:
            self._campaign = campaign
        
        if dataera is None:
            raise Exception('Unable to determine data era of JECs!')
        else:
            self._dataera = dataera
        
        if datatype is None:
            raise Exception('Unable to determine if JECs are for MC or Data!')
        else:
            self._datatype = datatype
        
        if len(levels) == 0:
            raise Exception('No levels provided?')
        else:
            self._levels = levels
            self._funcs = funcs
        
        if jettype is None:
            raise Exception('Unable to determine type of jet to correct!')
        else:
            self._jettype = jettype
        
        for i,level in enumerate(self._levels):
            this_level = _getLevel(level)
            ord_idx = _level_order.index(this_level)
            if i != this_level:
                self._levels[i],self._levels[ord_idx] = self._levels[ord_idx],self._levels[i]
                self._funcs[i],self._funcs[ord_idx] = self._funcs[ord_idx],self._funcs[i]
    
        #now we setup the call signature for this factorized JEC
        self._signature = []
        for func in self._funcs:
            sig = func.signature
            for input in sig:
                if input not in self._signature:
                    self._signature.append(input)

    @property
    def signature(self):
        """ list the necessary jet properties that must be input to this function """
        return self._signature
    
    def __repr__(self):
        out  = 'campaign   : %s\n'%(self._campaign)
        out += 'data era   : %s\n'%(self._dataera)
        out += 'data type  : %s\n'%(self._datatype)
        out += 'jet type   : %s\n'%(self._jettype)
        out += 'levels     : %s\n'%(','.join(self._levels))
        out += 'signature  : (%s)\n'%(','.join(self._signature))
        return out
    
    def getResolution(self,**kwargs):
        """
            Returns the set of resolutions for all input jets at the highest available level
            use like:
            jrs = reso.getResolution(JetProperty1=jet.property1,...)
            'jrs' will be formatted like [[jr_jet1 jr_jet2 ...] ...]
        """
        resos = []
        for i,func in enumerate(self._funcs):
            sig = func.signature
            args = []
            for input in sig:
                args.append(kwargs[input])
            resos.append(func(*tuple(args)))                        
        return resos[-1]



# here we want to define a group of columns from striped
# that all pertain to the same object type
# i.e. group = PhysicalColumnGroup(job,p4="Electron_p4",""NElectrons","Electron_p4","Electron_id1",...)
# this is specialized to striped so far
# we should try to peel it apart for other things

class ColumnGroup(object):
    def __init__(self,events,objName,*args):
        self._map = {}
        eventObj = getattr(events,objName)
        self._counts = getattr(events,objName).count
        for arg in args:
            callStack = arg.split('.')
            retval = getattr(eventObj,callStack[0])
            for i in range(1,len(callStack)):
                retval = getattr(retval,callStack[i])
            self._map[arg] = retval
            
    def __getitem__(self,name):
        return self._map[name]
    
    def columnsWithout(self,toremove):
        out = {}
        out.update(self._map)
        if isinstance(toremove,str):
            del out[toremove]
        else:
            for key in toremove:
                del out[key]
        return out
    
    def columns(self):
        return self._map
    
    def counts(self):
        return self._counts
    
class PhysicalColumnGroup(ColumnGroup):
    def __init__(self,events,objName,*args,**kwargs):
        self._hasp4 = False
        argkeys = kwargs.keys()
        self._p4 = None
        if 'p4' in argkeys:
            self._hasp4 = True
            self._p4 = {'p4':kwargs['p4']}
        elif 'pt' in argkeys and 'eta' in argkeys and 'phi' in argkeys and 'mass' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['pt','eta','phi','mass']}
        elif 'pt' in argkeys and 'eta' in argkeys and 'phi' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['pt','eta','phi','energy']}
        elif 'px' in argkeys and 'py' in argkeys and 'pz' in argkeys and 'mass' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['px','py','pz','mass']}
        elif 'pt' in argkeys and 'phi' in argkeys and 'pz' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['pt','phi','pz','energy']}
        elif 'px' in argkeys and 'py' in argkeys and 'pz' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['px','py','pz','energy']}
        elif 'p' in argkeys and 'theta' in argkeys and 'phi' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['p','theta','phi','energy']}
        elif 'p3' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key:kwargs[key] for key in ['p3','energy']}
        else:
            raise Exception('No valid definition of four-momentum found to build JaggedCandidateArray')
        allargs = list(args) + [val for val in kwargs.values()]
        super(PhysicalColumnGroup,self).__init__(events,objName,*allargs)
        
    def p4Name(self):
        if self._p4 is None:
            raise Exception('p4 is not set for this PhysicalColumnGroup')
        return self._p4.values()
    
    def p4Columns(self):
        return {key:self.columns()[value] for key,value in self._p4.items()}
    
    def otherColumns(self):
        return self.columnsWithout(self.p4Name())




from .hist_tools import SparseAxis, DenseAxis
from uproot_methods.classes.TH1 import Methods as TH1Methods


def export1d(hist, axis):
    if hist.dense_dim() != 1:
        raise ValueError("export1d() can only support one dense dimension")

    axis = hist.axis(axis)
    if not isinstance(axis, DenseAxis):
        raise ValueError("export1d() only exports dense axes")

    sumw, sumw2 = hist.values(sumw2=True, overflow='all')[()]
    edges = axis.edges(overflow='none')

    class TH1(TH1Methods, list):
        pass

    class TAxis(object):
        def __init__(self, fNbins, fXmin, fXmax):
            self._fNbins = fNbins
            self._fXmin = fXmin
            self._fXmax = fXmax

    out = TH1.__new__(TH1)
    out._fXaxis = TAxis(len(edges) - 1, edges[0], edges[-1])
    out._fXaxis._fName = axis.name
    out._fXaxis._fTitle = axis.label
    if not axis._uniform:
        out._fXaxis._fXbins = edges.astype(">f8")

    centers = (edges[:-1] + edges[1:]) / 2.0
    out._fEntries = out._fTsumw = out._fTsumw2 = sumw[1:-1].sum()
    out._fTsumwx = (sumw[1:-1] * centers).sum()
    out._fTsumwx2 = (sumw[1:-1] * centers**2).sum()

    out._fName = "histogram"
    out._fTitle = hist.label

    out._classname = b"TH1D"
    out.extend(sumw.astype(">f8"))
    out._fSumw2 = sumw2.astype(">f8")

    return out

